-module(rek).
-compile(export_all).
-export([sum/1, sumi/1, fib/1, fibi/1, test/0, m/1]).
-author('kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2015-11-02 20:02:18 +0100 (h, 02 nov 2015) $$').

sum([])    -> print_mem_usage("Rekurzio melyen"), 0; % rekurzió mélyén kiírjuk a memóriát
sum([H|T]) -> H + sum(T).

% @type btree() = leaf | {any(),btree(),btree()}.
%leaves(leaf)      -> 1;
%leaves({_,Lt,Rt}) -> leaves(Lt) + leaves(Rt).

sumi(L) -> sumi(L,0).

% sumi(L,N) az L lista összege + N.
sumi([], N)    -> print_mem_usage("Rekurzio melyen"), N; % rekurzió mélyén kiírjuk a memóriát
sumi([H|T], N) -> sumi(T,N+H).

fib(0) -> 0;
fib(1) -> 1;
fib(N) -> fib(N-2) + fib(N-1).

% fibi(N) az N. Fibonacci-szám.
fibi(0) -> 0;
fibi(N) -> fibi(N-1, 1, 0).

% fibi(N,A,B) az  A<-A+B, B<-A  transzformáció N-szeri ismétlése utáni A.
fibi(0, A, _B) -> A;
fibi(I, A, B)  -> fibi(I-1, B+A, A).

% @spec map(fun(A) -> B, [A]) -> [B].
map(_F, [])    -> [];
map(F, [X|Xs]) -> [F(X)|map(F, Xs)].

append([], L2) -> print_mem_usage("Rekurzio melyen"), L2;
append(L1, L2) -> [hd(L1)|append(tl(L1), L2)].

revapp([], L2) -> print_mem_usage("Rekurzio melyen"), L2;
revapp(L1, L2) -> revapp(tl(L1), [hd(L1)|L2]).


test() ->
    {
      [ rek:fib(N) || N <- [0,1,2,3,4,5]],
      [ rek:fibi(N) || N <- [0,1,2,3,4,5]],
      sum([1,2,3]),
      sumi([1,2,3]),
      tree:leaves({1,{2,{3,leaf,leaf},
                      leaf},
                   {5,leaf,leaf}}),
      map(fun(X) -> X+1 end, [10,20,30])
    }.

% Durva becslés: kiírja a függvényhívás előtt és után használt memóriát.
print_mem_usage(Msg) ->
%    [{total,T1}|_] = erlang:memory(),
    {memory,M1} = erlang:process_info(self(), memory),
    {_,Time} = statistics(runtime),
%    io:format("~16s: ~3w M / ~3w M  |  ~.3f s~n", [Msg, M = M1 div 1024 div 1024, T = T1 div 1024 div 1024, Time / 1000]),
    io:format("~16s: ~3w M  |  ~.3f s~n", [Msg, M = M1 div 1024 div 1024, Time / 1000]),
%    io:format("~p~n", [erlang:memory()]),
    {M,Time}.

memtest1(Fun) ->
    process_flag(trap_exit, true),
    L1 = lists:seq(1,1500000),
    Pid = spawn_link(
      fun() ->
              print_mem_usage("Hivas elott"),
              S = Fun(L1),
              print_mem_usage("Hivas utan"),
              timer:sleep(1000),
              exit(S)
      end),
    receive {'EXIT', Pid, R} -> ok end,
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    R.
    
memtest2(Fun) ->
    process_flag(trap_exit, true),
    L1 = lists:seq(1,1500000),
    L2 = [-1],
    Pid = spawn_link(
      fun() ->
              print_mem_usage("Hivas elott"),
              L3 = Fun(L1,L2),
              print_mem_usage("Hivas utan"),
              timer:sleep(1000),
              exit(length(L3))
      end),
    receive {'EXIT', Pid, R} -> ok end,
    erlang:garbage_collect(),timer:sleep(1000),statistics(runtime),
    R.

m(1) -> memtest1(fun sum/1);
m(2) -> memtest1(fun sumi/1);
m(3) -> memtest1(fun(L) -> lists:foldr(fun erlang:'+'/2, 0, L) end);
m(4) -> memtest1(fun(L) -> lists:foldl(fun erlang:'+'/2, 0, L) end);
m(5) -> memtest2(fun append/2);
m(6) -> memtest2(fun lists:append/2);
m(7) -> memtest2(fun erlang:'++'/2);
m(8) -> memtest2(fun revapp/2);
m(9) -> memtest2(fun lists:reverse/2);
m(10) -> memtest2(fun(L1,L2) ->
                          R1=lists:reverse(L1),
                          print_mem_usage("Kozben"),
                          erlang:garbage_collect(),timer:sleep(2000),
                          print_mem_usage("Kozben GC utan"),
                          revapp(R1,L2)
                  end);
m(11) -> memtest2(fun(L1,L2) ->
                          R1=revapp(L1, []),
                          print_mem_usage("Kozben"),
                          erlang:garbage_collect(),timer:sleep(2000),
                          print_mem_usage("Kozben GC utan"),
                          revapp(R1,L2)
                  end).

t() ->
    [ begin
          io:format("m(~p)~n", [I]),
          print_mem_usage("Kezdetben"),
          m(I)
      end
      || I<-[1,2,3,4,5,6,7,8,9,10,11] ].

-ifdef(komment).

%% % Kimenet:
%% 2> rek:t().
%% m(1)                                            % sum
%%        Kezdetben:   0 M  |  0.050 s
%%      Hivas elott: 233 M  |  1.080 s
%%  Rekurzio melyen: 583 M  |  0.390 s
%%       Hivas utan: 583 M  |  0.070 s
%% m(2)                                            % sumi
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  1.060 s
%%  Rekurzio melyen: 233 M  |  0.120 s
%%       Hivas utan: 233 M  |  0.000 s
%% m(3)                                            % foldr(+/2)
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  1.060 s
%%       Hivas utan: 570 M  |  1.020 s
%% m(4)                                            % foldl(+/2)
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  1.020 s
%%       Hivas utan: 233 M  |  0.380 s
%% m(5)                                            % append
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  1.000 s
%%  Rekurzio melyen: 583 M  |  0.560 s
%%       Hivas utan: 583 M  |  0.060 s
%% m(6)                                            % lists:append
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  0.980 s
%%       Hivas utan: 570 M  |  0.140 s
%% m(7)                                            % ++
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  0.960 s
%%       Hivas utan: 570 M  |  0.170 s
%% m(8)                                            % revapp
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  0.990 s
%%  Rekurzio melyen: 598 M  |  0.540 s
%%       Hivas utan: 598 M  |  0.000 s
%% m(9)                                            % lists:reverse
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  1.070 s
%%       Hivas utan: 598 M  |  0.180 s
%% m(10)                                           % revapp(reverse)
%%        Kezdetben:   0 M  |  0.000 s
%%      Hivas elott: 233 M  |  0.910 s
%%           Kozben: 598 M  |  0.190 s
%%  Rekurzio melyen: 456 M  |  0.380 s
%%       Hivas utan: 456 M  |  0.000 s

-endif.
