-module(set).
-compile(export_all).
-author('hanak@iit.bme.hu, modified by: kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2011-12-01 20:57:09 +0100 (Thu, 01 Dec 2011) $$').

% @type set() = list().

% empty() az üres halmaz.
empty() ->
    [].

% isMember(X, S) igaz, ha az X elem benne van az S halmazban.
isMember(_, []) ->
    false;
isMember(X, [Y|Ys]) ->
    X=:=Y orelse isMember(X, Ys).

isMember2(_, []) ->
    false;
isMember2(X, [X|_]) ->
    true;
isMember2(X, [_Y|Ys]) ->
    isMember2(X, Ys).

% @spec newMember(X::any(), S1::set()) -> S2::set().
% S2 halmaz az S1 halmaz és az [X] halmaz uniója.
newMember(X, Xs) ->
    case isMember(X, Xs) of
        true -> Xs;
        false -> [X|Xs]
    end.

% @spec listToSet(list()) -> set().
% listToSet(L) az L lista elemeinek halmaza.
listToSet([]) ->
    [];
listToSet([X|Xs]) ->
    newMember(X, listToSet(Xs)).

% union(S1, S2) halmaz az S1 és S2 halmazok uniója.
union([], Ys) ->
    Ys;
union([X|Xs], Ys) ->
    newMember(X, union(Xs, Ys)).

union2(Xs, Ys) ->
    lists:foldr(fun newMember/2, Ys, Xs).

union3(Xs, Ys) ->
    lists:foldl(fun newMember/2, Ys, Xs).

% intersect(S1, S2) halmaz az S1 és S2 halmazok metszete.
intersect([], _) ->
    [];
intersect([X|Xs], Ys) ->
    Zs = intersect(Xs, Ys),
    case isMember(X, Ys) of
        true ->
            [X|Zs];
        false ->
            Zs
    end.

intersect2([], _) ->
    [];
intersect2([X|Xs], Ys) ->
    case isMember(X, Ys) of
        true ->
            [X|intersect2(Xs, Ys)];
        false ->
            intersect2(Xs, Ys)
    end.

intersect3(Xs, Ys) ->
    [ X || X <- Xs, isMember(X, Ys) ].

% isSubset(S1, S2) igaz, ha az S1 halmaz az S2 halmaz részhalmaza.
isSubset([], _) ->
    true;
isSubset([X|Xs], Ys) ->
    isMember(X, Ys) andalso isSubset(Xs, Ys).

% isEqual(S1, S2) igaz, ha az S1 halmaz és az S2 halmaz azonos.
isEqual(Xs, Ys) ->
    isSubset(Xs, Ys) andalso isSubset(Ys, Xs).

% powerSet1(S) az S halmaz hatványhalmaza.
powerSet1([]) ->
    [[]];
powerSet1([X|Xs]) ->
    P = powerSet1(Xs),
    P ++ [ [X|Ys] || Ys <- P ].

powerSet2(Xs) ->
    lists:foldl(fun(X, Xss) ->
                        Xss ++ [ [X|Ys] || Ys <- Xss ]
                end,
                [[]],
                Xs).

    
%% % pws(S1,Zs) mindazon halmazok halmaza, amelyek előállnak az S1 halmaz egy
%% % részhalmazának és a Zs halmaznak az uniójaként.
%% pws([],Zs) ->
%%     [Zs];
%% pws([X|Xs],Zs) ->
%%     pws(Xs,Zs) ++ pws(Xs,[X|Zs]).

%% % powerSet1(S) az S halmaz hatványhalmaza.
%% powerSet1(Xs) ->
%%     pws(Xs,[]).

% @spec insAll(X::any(), Yss::[[any()]], Zss::[[any()]]) -> Xss::[[any()]].
% Xss az Yss lista Ys elemeinek Zss elé fűzött
% listája, amelyben minden Ys elem elé X van beszúrva.
insAll(_X,[],Zss) ->
    Zss;
insAll(X,[Ys|Yss],Zss) ->
    insAll(X,Yss,[[X|Ys]|Zss]).

%% insAll2(X,Yss) ->
%%     [[X|Ys] || Ys <- Yss].

%% %% lineáris-rekurzív processzt elõállító változat
%% powerSet21([]) ->
%%     [[]];
%% powerSet21([X|Xs]) ->
%%     Pws = powerSet21(Xs),
%%     Pws ++ insAll(X,Pws,[]).

powerSet3([]) ->
    [[]];
powerSet3([X|Xs]) ->
    P = powerSet3(Xs),
    insAll(X,P,P).

%% powerSet23([]) ->
%%     [[]];
%% powerSet23([X|Xs]) ->
%%     Pws = powerSet23(Xs),
%%     Pws ++ insAll2(X,Pws).

%%% Példák

stopper(F, A) ->
    erlang:garbage_collect(),timer:sleep(100),
    _ = statistics(runtime),
    R=F(A),
    {length(R),element(2,statistics(runtime))}.

test() ->
    [
     {isMember,isMember(f,[1,5,g,v,6,7,d,c,y,q,w,",","=",alma,f,akna])},
     {isMember2,isMember2(a,[1,5,g,v,6,7,d,c,y,q,w,",","=",alma,akna])},
     {newMember,newMember(5,[0,2,4,6,8])},
     {newMember,newMember(5,[1,3,5,7,9])},
     {listToSet,listToSet([1,2,3,4,5,6,7,8,9,8,6,4])},
     {union,union([6,4,2],[9,7,3,5])},
     {union,union([7,6,4,3,2],[9,7,3,5,6])},
     %% {union2,union2([6,4,2],[9,7,3,5])},
     %% {union2,union2([7,6,4,3,2],[9,7,3,5,6])},
     %% {union3,union3([6,4,2],[9,7,3,5])},
     %% {union3,union3([7,6,4,3,2],[9,7,3,5,6])},
     {intersect,intersect([6,4,2],[9,7,3,5])},
     {intersect,intersect([7,6,4,3,2],[9,7,3,5,6])},
     %% {intersect2,intersect2([6,4,2],[9,7,3,5])},
     %% {intersect2,intersect2([7,6,4,3,2],[9,7,3,5,6])},
     %% {intersect3,intersect3([6,4,2],[9,7,3,5])},
     %% {intersect3,intersect3([7,6,4,3,2],[9,7,3,5,6])},
     {isSubset,isSubset([2,7,6,3],[9,7,3,5,6])},
     {isSubset,isSubset([7,6,3],[9,7,3,5,6])},
     {isSubset,isSubset([9,7,3,5,6],[7,6,3])},
     {isEqual,isEqual([7,6,3],[6,3,7])},
     {isEqual,isEqual([7,6,3],[7,6,3])},
     {isEqual,isEqual([7,6,3],[7,6,3,0])},
     {powerSet1,powerSet1([1,2,3])},
     {length,length(powerSet1(lists:seq(1,10)))},
     {powerSet2,powerSet2([1,2,3])},
     {length,length(powerSet2(lists:seq(1,10)))},
     {powerSet3,powerSet3([1,2,3])},
     {length,length(powerSet3(lists:seq(1,10)))},
     begin
         Canon = fun(PS) -> lists:usort([lists:usort(S) || S <- PS(lists:seq(1,10))]) end,
         Canon(fun set:powerSet1/1) =:= Canon(fun set:powerSet3/1)
     end
%     lists:usort([lists:usort(S) || S <- set:powerSet(lists:seq(1,10))]) =:= lists:usort([lists:usort(S) || S <- set:powerSet3(lists:seq(1,10))])
    ].


%% tp(0) ->
%%   powerSet1([1,2,3]);
%% tp(1) ->
%%   powerSet21([1,2,3]);
%% tp(2) ->
%%   powerSet22([1,2,3]);
%% tp(10) ->
%%   powerSet1([1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,19,20]);
%% tp(11) ->
%%   powerSet1([1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,19,20,21,22]);
%% tp(12) ->
%%   powerSet21([1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,19,20,21,22]);
%% tp(13) ->
%%   powerSet22([1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,19,20,21,22]);

